#!/bin/bash

#
# objrename.sh
#
# Neil Klingensmith June 6, 2017
#
# Renames symbols in the src directory. Useful for having two copies of the same
# RTOS distribution compiled into the same binary image.
#

APPENDSTR="\_g1"

# List the C and H files to be updated
flist=($(find . -type f \( -name "*.h" -or -name "*.c" \)))

for objfile in `find ../ -iname '*.o'`
do
    # Strip .o extension
    bn=`echo $objfile | cut -d'.' -f1`

    # Strip Debug directory from beginning of the path
    cpath=`echo $bn | cut -d'/' -f2-`

    # Append .c to the end of the filename
    cfile="$cpath.c"

    # Check if the object file came from the hermes directory. If so, skip it.
    # We don't need to rename symbols in the hermes distribution.
    ishermescode=`echo $objfile | grep hermes`;
    if [ ! -z "$ishermescode" ]
    then
        continue
    fi

    # Search for function names in the output of objdump
    echo "Function names in $objfile:"
    while read line
    do
        echo $line

        for ((i=0; i<${#flist[@]}; i++)); do

            # Check if the object file came from the hermes directory. If so, skip it.
            # We don't need to rename symbols in the hermes distribution.
            ishermescode=`echo $objfile | grep hermes`;
            if [ ! -z "$ishermescode" ]
            then
                echo "Skipping"
                continue
            fi

            sed -i "s/\b$line\b/$line$APPENDSTR/g" ${flist[$i]}

        done

    done < <(objdump --syms $objfile | grep text | rev | cut -f1 -d' ' | rev | grep -v ".text" | grep -v -e '^$' | grep -v '\$')

    # Search for variables in the output of objdump
    echo "Variable names in $objfile:"
    while read line
    do
        echo $line

        for ((i=0; i<${#flist[@]}; i++)); do

            # Check if the object file came from the hermes directory. If so, skip it.
            # We don't need to rename symbols in the hermes distribution.
            ishermescode=`echo $objfile | grep hermes`;
            if [ ! -z "$ishermescode" ]
            then
                echo "Skipping"
                continue
            fi

            sed -i "s/\b$line\b/$line$APPENDSTR/g" ${flist[$i]}

        done

    done < <(objdump --syms $objfile | grep 'bss\|data\|rodata\|COM' | rev | cut -f1 -d' ' | rev | grep -v ".data\|.bss\|.rodata\|*COM*" | grep -v "." | grep -v -e '^$' | grep -v '\$')

done


